/* jigsaver.c by bill buckels 1991 */
/* a jigsaw screensaver            */
/* written in microsoft C 5.1 Large Model */
/* CGA compatible displays only and BSAVED and CGA PCX images only... */
/* used bios routines and teletype input... functional. */


#include <malloc.h>
#include <dos.h>
#include <fcntl.h>
#include <stdio.h>
#include <bios.h>
#include <string.h>

/* buffer stuff for the puzzle */
char far *jigpieces[40];         /* 40 buffers for the pieces  */
#define PIECE_SIZE 400           /* 400 bytes per piece        */

/* logical grouping for CGA mode graphics jigsaw puzzle */
/* dimensions are X1, Y1, X2, Y2                        */
/* dimensions are BYTE ORIENTED                         */

int byte_puzzle[40][4]={

              0,    0,  9, 39, /* groups of 8 horizontally */
              10,   0, 19, 39, /* groups of 5 vertically   */
              20,   0, 29, 39,
              30,   0, 39, 39,
              40,   0, 49, 39,
              50,   0, 59, 39,
              60,   0, 69, 39,
              70,   0, 79, 39,

              0,   40,  9, 79, /* groups of 8 horizontally */
              10,  40, 19, 79, /* groups of 5 vertically   */
              20,  40, 29, 79,
              30,  40, 39, 79,
              40,  40, 49, 79,
              50,  40, 59, 79,
              60,  40, 69, 79,
              70,  40, 79, 79,

              0,   80,  9,119, /* groups of 8 horizontally */
              10,  80, 19,119, /* groups of 5 vertically   */
              20,  80, 29,119,
              30,  80, 39,119,
              40,  80, 49,119,
              50,  80, 59,119,
              60,  80, 69,119,
              70,  80, 79,119,

              0,  120,  9,159, /* groups of 8 horizontally */
              10, 120, 19,159, /* groups of 5 vertically   */
              20, 120, 29,159,
              30, 120, 39,159,
              40, 120, 49,159,
              50, 120, 59,159,
              60, 120, 69,159,
              70, 120, 79,159,

              0,  160,  9,199, /* groups of 8 horizontally */
              10, 160, 19,199, /* groups of 5 vertically   */
              20, 160, 29,199,
              30, 160, 39,199,
              40, 160, 49,199,
              50, 160, 59,199,
              60, 160, 69,199,
              70, 160, 79,199};

/* the indices for each piece */
int newposition[40];
#define NUM_PIECES 40


/* pointers to the screen */
unsigned char far *crt    =(unsigned char *) 0xB8000000l;
unsigned char far *inleaf =(unsigned char *) 0xB8000000l+0x2000;
#define CGALINE 80

void lode(int x1,int y1,int x2, int y2,int position)
{
    int inset=0;
    int offset=((y1/2)*CGALINE)+x1;
    unsigned int height=(y2-y1)+1;
    unsigned int width= (x2-x1)+1;
    int limit=y2+1;

    while(y1<limit)
    {
    memcpy(&crt[offset],&jigpieces[position][inset],width);
    inset+=(width);
    y1++;
    if(y1<limit)
    memcpy(&inleaf[offset],&jigpieces[position][inset],width);
    inset+=(width);
    offset+=CGALINE;
    y1++;
    }

}

void unlode(int x1,int y1,int x2, int y2,int position)
{
    int inset=0;
    int offset=((y1/2)*CGALINE)+x1;
    unsigned int height=(y2-y1)+1;
    unsigned int width= (x2-x1)+1;
    int limit=y2+1;

    while(y1<limit)
    {
    memcpy(&jigpieces[position][inset],&crt[offset],width);
    inset+=(width);
    y1++;
    if(y1<limit)
    memcpy(&jigpieces[position][inset],&inleaf[offset],width)  ;
    inset+=(width);
    offset+=CGALINE;
    y1++;
    }

}


void setup(void)
{
  int i=0;
  while(i<NUM_PIECES)
  {
   jigpieces[i]=malloc(PIECE_SIZE);
   newposition[i]=i;
   i++;
  }

}

void setdown(void)
{
  int i=0;
  while(i<NUM_PIECES)
  {
   free(jigpieces[i]);
   i++;
  }
}



int setwindow(unsigned r1, unsigned c1, unsigned r2, unsigned c2)
{
    union REGS inregs, outregs;

    inregs.h.ah =6;
    inregs.h.al =0;
    inregs.h.bh =0;
    inregs.h.cl =c1;
    inregs.h.ch =r1;
    inregs.h.dl =c2;
    inregs.h.dh =r2;
    int86(0x10,&inregs,&outregs);

    return 0;

}

int locate(unsigned row, unsigned col)
{
    union REGS inregs, outregs;

    /* position cursor */
    inregs.h.ah =2;
    inregs.h.bh =0;
    inregs.h.dh =row;
    inregs.h.dl =col;
    int86(0x10,&inregs,&outregs);

}


int midplots(char *str, unsigned row, unsigned width)
{
    union REGS inregs, outregs;
    unsigned col=(width-strlen(str))/2;

    /* position cursor */
    locate(row,col);
    /* print the string */
    printf("%s",str);
    return col;

}

int setcrtmode(unsigned char _CRT_MODE)
{
    union REGS rin,rout;

    rin.h.ah = 0;
    rin.h.al = _CRT_MODE;
    int86(0x10,&rin,&rout);


    rin.h.ah= 0x0f;
    int86(0x10,&rin,&rout);
    if(rout.h.al!=_CRT_MODE)
    {
    printf(
    "SORRY... VIDEO MODE %d is required but NOT supported by YOUR ADAPTER.\n",
    _CRT_MODE);
    exit(0);
    }
    return 0;

}


/* type conversion function */
unsigned int byteword(unsigned char a, unsigned char b){return b<<8|a;}
unsigned screen_mode = 5;

int checkforpcx(char *pcxheader)
{
    int i;    /* we only want CGA compatible full screens.           */

    unsigned int zsoft,codetype,pixbits;
    unsigned int hres, vres;
    unsigned int no_planes, bytesperline;
    int invalid = -1, valid = 0, status=valid;

    zsoft   =pcxheader[0];
    codetype=pcxheader[2];
    pixbits =pcxheader[3];

    if(zsoft!=10)        status = invalid;
    if(codetype!=1)      status = invalid;
    if(pixbits != 2 )
       if(pixbits != 1)  status = invalid;

    hres=byteword(pcxheader[12],pcxheader[13]);
    vres=byteword(pcxheader[14],pcxheader[15]);
    no_planes   =pcxheader[65];
    bytesperline=byteword(pcxheader[66],pcxheader[67]);


    if(hres != 320){
       screen_mode=6;
       if(hres!=640)     status = invalid;
       else setcrtmode(6);
       }
    if(vres != 200)      status = invalid;
    if(no_planes!=1)     status = invalid;
    if(bytesperline !=80)status = invalid;
    /* we can ignore the color map since we        */
    /* are limiting ourselves to CGA modes         */
    /* so we will not handle over 2-bits per pixel */
    return status;

}


int pcxload(char *pcxfilename)
{
    unsigned int byteoff=0,leaf=8192,packet,width=0;
    FILE *fp;
    unsigned char byte,bytecount;
    long wordcount,target;
    char pcxheader[128];
    char *wordptr;


    fp = fopen(pcxfilename,"rb");
    target = filelength(fileno(fp));
    for(wordcount=0;wordcount!=128;wordcount++)pcxheader[wordcount]=fgetc(fp);
    if(checkforpcx(pcxheader)==-1)
    {
        fclose(fp);
        return-1;
    }


    do{ bytecount=1;                          /* start with a seed count */
        byte=fgetc(fp);
        wordcount++;
                                              /* check to see if its raw */
        if(0xC0 == (0xC0 &byte)){             /* if its not, run encoded */
                    bytecount= 0x3f &byte;
                    byte=fgetc(fp);
                    wordcount++;
                    }
        for(packet=0;packet<bytecount;packet++){
                     if(width<80){ crt[byteoff]=byte;
                                   width++;
                                   byteoff++;
                                   }
                     else{crt[leaf]=byte;
                                   leaf++;
                                   width++;
                                   if(width>159)width=0;
                                   }
                                   }
        }while(wordcount<target);
        fclose(fp);
        return(0);
}




int bload(char *filebuf)
{
       FILE *fp;
       int fh;
       unsigned row,col=80;

       char bfile[66];
       char pfile[66];
       char messfile[66];

       char headbuf[7];
       char messbuf[66];
       char scratch[66];

       char *wordptr;

       strcpy(scratch,filebuf);
       wordptr=strtok(scratch,".");

       sprintf(messfile,"%s.TXT",scratch);
       sprintf(bfile,"%s.BAS",scratch);
       sprintf(pfile,"%s.PCX",scratch);



       if((fh=open(bfile,O_RDONLY|O_BINARY))!=-1)
       {
         read(fh,headbuf,7);
         if(headbuf[0]!='\xfd')
         {
            close(fh);
            return -1;
         }
         read(fh,crt,0x4000);
         close(fh);
       }
       else if(pcxload(pfile)==-1)return -1;

      if((fp=fopen(messfile,"r"))!=NULL)
         {
         row=25;
         if(screen_mode!=6)col=40;
         while(fgets(messbuf,128,fp)!=NULL)row--;
         fclose(fp);
         setwindow(row,0,24,col-1);
         fp=fopen(messfile,"r");
         while(fgets(messbuf,128,fp)!=NULL)
         {
          wordptr=strtok(messbuf,"\n");
          midplots(messbuf,row,col);
          row++;
          }
        fclose(fp);
        }

        return 0;

}


int random(int seed)
{
  int hold=-1;           /* returns a random number in a usable range */
  while(hold<0)hold=rand();
  while(hold >=seed)hold-=seed;
  return hold;
}



/* save the screen into arrays */
void getpieces()
{
    int i=0;
    while(i<NUM_PIECES)
    {
     unlode(byte_puzzle[i][0],byte_puzzle[i][1],
            byte_puzzle[i][2],byte_puzzle[i][3],
            i);
           i++;
    }
}

/* put the shuffled arrays onto the screen */
void shuffledpieces()
{
    int i=0;
    while(i<NUM_PIECES)
    {
       lode(byte_puzzle[i][0],byte_puzzle[i][1],
            byte_puzzle[i][2],byte_puzzle[i][3],
            newposition[i]);
           i++;
    }



}


/* some interrupt stuff for the screen saver cycle */

int setinterrupt(unsigned char intno, unsigned segment, unsigned offset)
{
    union REGS inregs, outregs;
    struct SREGS segregs;

    inregs.h.ah = 0x25                    ; /* msdos function 25h         */
    inregs.h.al = intno                   ; /* interrupt number requested */
    segregs.ds  = segment                 ; /* segment of address pointer */
    inregs.x.dx = offset                  ; /* offset of address pointer  */
    int86x(0x21,&inregs,&outregs,&segregs); /* generate MSDOS interrupt   */

    return 0;
}


int getinterrupt(unsigned char intno, unsigned *segment, unsigned *offset)
{
    union REGS inregs, outregs;
    struct SREGS segregs;

    inregs.h.ah = 0x35                    ; /* msdos function 35h         */
    inregs.h.al = intno                   ; /* interrupt number requested */
    int86x(0x21,&inregs,&outregs,&segregs); /* generate MSDOS interrupt   */
    *segment = segregs.es                 ; /* store the segment pointer  */
    *offset  = outregs.x.bx               ; /* store the offset pointer   */

    return 0;
}






/* our replacement for whatever was hooked to the DOS timer tick */
/* a simple strobe */
unsigned ticker = 0;
void interrupt timertick()
{
    ticker++;
}


int clearblue()
{
   int i=0;
   char *ptr;
   ptr=(char *)&crt[0];

   while(i<2000)
   {
     if(i%80==0 || i%80==79 || i<80 || i>(2000-80))*ptr++=219;
     else *ptr++=32;
     *ptr++=0x1f;
     i++;
    }

}

main(int argc, char **argv)
{


    FILE *fp;
    unsigned segment, offset;
    int position, oldposition, temp,i;
    char filebuf[66];
    char menuitems[48][40];
    int ctr;
    char *wordptr;


    char c; /* seed the keypress query character */
    float target=(float)18.2*10; /* default to 10 seconds for screensaver */


    setcrtmode(3);

REDO:;

    ctr=0;
    c='\x01';
    ticker=0;
    screen_mode=5;

    if(argc==2)strcpy(filebuf,argv[1]);
    else
    {
    clearblue();
    if((fp=fopen("PICLIST","r"))==NULL)
    {
    midplots(" JIGSAVER(C) by Bill Buckels 1991 ",0,80);
    midplots("CLIPBOARD-SCREENSAVER for CGA or Compatible Displays.",11,80);
    midplots("USAGE is \"JIGSAVER [BSAVED.BAS]\"",12,80);
    midplots(
    "A TEXT MESSAGE of up to 40 words may also be included...",13,80);
    midplots(
    "The Message File is in ASCII format i.e. \"[BSAVED.TXT]\"",14,80);
    midplots(" Enter A Blank Line To Exit ",24,80);
    midplots("Enter The FILENAME : ",6,80);
    gets(filebuf);
    }
    else
    {
       midplots(" JIGSAVER(C) by Bill Buckels 1991 ",0,80);
        midplots(
      " Select From The Choices Above ** or ** A Blank Line To Exit ",24,80);

        while(fgets(filebuf,128,fp)!=NULL)
        {
         wordptr=strtok(filebuf,"\n");
         if(ctr<44)
         {
            sprintf(menuitems[ctr],"%s",filebuf);
            ctr++;
            }
        }
        fclose(fp);

        if(ctr<22)i=(25-ctr)/2;
        else i=(50-ctr)/4;

        for(temp=0;temp<ctr;temp++)
        {
         sprintf(filebuf,"%-2d- %15s",temp+1,menuitems[temp]);
         if(ctr<22)position=midplots(filebuf,i,80);
         else
         {
          position=midplots(filebuf,i,40);
          temp++;
          if(temp<ctr)
          {
            sprintf(filebuf,"%-2d- %15s",temp+1,menuitems[temp]);
            position=midplots(filebuf,i,120);
            }

         }
         i++;
        }
        locate(i,position);
        gets(filebuf);
        temp=atoi(filebuf);
        if(temp>0 && temp <(ctr+1))strcpy(filebuf,menuitems[temp-1]);

    }
    }
      setcrtmode(5);
      if(bload(filebuf)!=-1)
        {
        setup();
        getpieces();

        /* 1. show the picture */
        shuffledpieces();


       /* 2. keep busy...                      */
       /* hook a vector for a ticker...        */
       /* use it to strobe the screen shuffler */

        getinterrupt(0x1c,&segment,&offset);
       _disable();
       _dos_setvect(0x1C, timertick);
       _enable();

       do{
            if(kbhit())
            {
             c=getch();
             if(c==0)getch();
             switch(c)
             {
                case 27: break;
                default: ticker=0;
                         for(i=0;i<NUM_PIECES;i++)newposition[i]=i;
                         shuffledpieces();
                         break;
                }
              }

            /* change the screen once per second... */
            /* by swapping 2-pieces...              */
            if(ticker>target)
            {
                ticker-=18;
                oldposition=random(NUM_PIECES);
                position=random(NUM_PIECES);
                temp = newposition[position];
                newposition[position]=newposition[oldposition];
                newposition[oldposition]=temp;
                lode(byte_puzzle[position][0],byte_puzzle[position][1],
                     byte_puzzle[position][2],byte_puzzle[position][3],
                     newposition[position]);
               lode(byte_puzzle[oldposition][0],byte_puzzle[oldposition][1],
                     byte_puzzle[oldposition][2],byte_puzzle[oldposition][3],
                     newposition[oldposition]);
                }

            }while(c!=27);

        /* 3. now we are done */
        /* put back the timertick and clean-up before leaving... */
        setinterrupt(0x1c,segment,offset);
        setdown();
        setcrtmode(3);
        if(argc==1)goto REDO;
        }
        else
        {
             setcrtmode(3);
             printf("Could Not Display %s...\n",filebuf);
            }

    exit(0);

}






